--[[

	CTRA TankAssist v5.040
	by CorpsusVitale

	Helper for CT Raid Assist.  Shows warning boxes
	next to the MT windows if the MTs are on the same
	target (useful for fights like Garr, Ragnaros etc)

	Accepts a number of commands:

	/ctrata show - Shows the tank warning windows, if there are any warnings
	/ctrata hide - Hides the warning windows and disables processing.
	/ctrata toggle - Shows if hidden, hides if shown
	/ctrata position left - Shows the warning windows to the left of the CTRA MT windows
	/ctrata position right - Shows the warning windows to the right of the CTRA MT windows
	/ctrata position auto - Tries to put the warning windows on the side of the CTRA MT windows
							that has more room
	/ctrata debug - Toggle debugging, but will remember the last set debug level.  Uses 1 if not set
	/ctrata debug n - Set the debug level to n (an integer)
	/ctrata version - Shows the current version
]]                       

-- Strings
BINDING_NAME_CTRA_TA_TOGGLE = "Toggle windows";
BINDING_NAME_CTRA_TA_TOGGLE_DESCRIPTION = "Toggle displaying the CTRA Tank Assist windows";
BINDING_HEADER_CTRA_TA = "CTRA Tank Assist";
CTRA_TA_VERSION_TEXT = "CTRA Tank Assist v6.0";

-- Global debug setting (volatile)
ctrata_debuglevel = 0;

--[[
	============================================================
	ctrata_debug

	Print a debug message if the current debugging level permits
	it.  The message is printed to the chat log and to the
	error frame in the top middle of the screen.
]]
function ctrata_debug(level, text)
	if(ctrata_debuglevel >= level) then
		DEFAULT_CHAT_FRAME:AddMessage(text)
		UIErrorsFrame:AddMessage(text, 1.0, 1.0, 0, 1, 10)
	end;
end

--[[
	============================================================
	ctrata_chat

	Print a message to the chat log
]]
function ctrata_chat(text)
	DEFAULT_CHAT_FRAME:AddMessage(text);
end

--[[
	============================================================
	CTRA_TA_OnLoad

	OnLoad handler for our dummy frame.  Just prints a message,
	registers the slash command handler and registers for the
	variable loaded event.
]]
function CTRA_TA_OnLoad()
	-- Say hello
	ctrata_chat("|cc0ffff00" .. CTRA_TA_VERSION_TEXT .. "|r loaded (use /ctrata)");

    -- Register slash command
    SLASH_CTRA_TA1 = "/ctrata";
    SlashCmdList["CTRA_TA"] = CTRA_TA_SlashHandler;
                                          
	-- Register loaded event
	this:RegisterEvent("VARIABLES_LOADED");
end;

--[[
	============================================================
	CTRA_TA_OnEvent

	Only called when the saved variables for this mod are loaded.
	We do default the settings here if they were not present.

	Based on the saved setting, we do the show or hide of the main dummy window
	here.  This will trigger an OnUpdate call, and cause the other windows to
	be hidden or shown.

	We let the user know whether or not the windows will be shown, since
	it isn't likely that they will be in a raid when it loads.
]]
function CTRA_TA_OnEvent()
	if ( event == "VARIABLES_LOADED" ) then
		CTRA_TA_Options = CTRA_TA_Options or {showWindow=false, position="auto"};

		if (CTRA_TA_Options.showWindow == true) then
		    ctrata_chat("|cc0ffff00CTRA_TA|r - showing the CTRA_TA windows");
			CTRA_TA_Window:Show();
		else
		    ctrata_chat("|cc0ffff00CTRA_TA|r - not showing the CTRA_TA windows");
			CTRA_TA_Window:Hide();
		end;
	end;
end;

--[[
	============================================================
	CTRA_TA_SlashHandler

	Slash command handler for /ctrata.

	Accepts a number of commands:
	 -  /ctrata show - Shows the tank warning windows, if there are any warnings
	 -  /ctrata hide - Hides the warning windows and disables processing.
	 -  /ctrata toggle - Shows if hidden, hides if shown
	 -  /ctrata position left - Shows the warning windows to the left of the CTRA MT windows
	 -  /ctrata position right - Shows the warning windows to the right of the CTRA MT windows
	 -  /ctrata position auto - Tries to put the warning windows on the side of the CTRA MT windows
								that has more room
	 -  /ctrata debug - Toggle debugging, but will remember the last set debug level.  Uses 1 if not set
	 -  /ctrata debug n - Set the debug level to n (an integer)
	 -  /ctrata version - Shows the version number

]]
function CTRA_TA_SlashHandler(cmd)
    ctrata_debug(1, "CTRA_TA_SlashHandler: " .. cmd);
    if(cmd) then
	    if(cmd == "show") then
			CTRA_TA_Options.showWindow = true;
    	    CTRA_TA_Window:Show();
	    elseif(cmd == "toggle") then
			if (CTRA_TA_Window:IsShown()) then
				CTRA_TA_Options.showWindow = false;
	    	    CTRA_TA_Window:Hide();
			else
				CTRA_TA_Options.showWindow = true;
    	    	CTRA_TA_Window:Show();
			end;
    	elseif (cmd == "hide") then
			CTRA_TA_Options.showWindow = false;
    	    CTRA_TA_Window:Hide();
		elseif (string.find(cmd, "position ")) then
			local posstr = string.sub(cmd, 10);
			ctrata_debug(1, "Position is '"..posstr.."'");
			if (posstr == "left") then
				CTRA_TA_Options.position = "left";
				ctrata_chat("CTRA_TA - position set to left");
			elseif (posstr == "right") then
				CTRA_TA_Options.position = "right";
				ctrata_chat("CTRA_TA - position set to right");
			elseif (posstr == "auto") then
				CTRA_TA_Options.position = "auto";
				ctrata_chat("CTRA_TA - position set to auto");
			else
    			ctrata_chat("CTRA_TA - position must be one of 'left', 'right' or 'auto'");
			end;
	    elseif (string.find(cmd, "debug ")) then
			-- Set the debugging level directly
			ctrata_debuglevel = tonumber(string.sub(cmd, 7)) or ctrata_debuglevel;
			ctrata_chat("CTRA_TA - debug level set to "..ctrata_debuglevel);
		elseif (cmd == "debug") then
			-- Toggle debugging, but save the current level
			if (ctrata_debuglevel == 0) then
				ctrata_debuglevel = 1;
			else
				ctrata_debuglevel = -ctrata_debuglevel;
			end;
			ctrata_chat("CTRA_TA - debug level set to "..ctrata_debuglevel);
		elseif (cmd == "help" or cmd == "") then
	        -- Show help
	        CTRA_TA_Usage_Information();
		elseif (cmd == "version" or cmd == "ver") then
	        -- Show version
			ctrata_chat("|cc0ffff00" .. CTRA_TA_VERSION_TEXT);
	    end;
	end;
end;

--[[
	============================================================
	CTRA_TA_ToggleWindow

	Keybinding handler - Toggle the display of the windows. We
	add a chat message in case there aren't actually any windows
	currently visible.
]]
function CTRA_TA_ToggleWindow()
	if (CTRA_TA_Window:IsShown()) then
		CTRA_TA_Options.showWindow = false;
	    CTRA_TA_Window:Hide();
	    ctrata_chat("|cc0ffff00CTRA_TA|r - hiding the CTRA_TA windows");
	else
		CTRA_TA_Options.showWindow = true;
		CTRA_TA_Window:Show();
	    ctrata_chat("|cc0ffff00CTRA_TA|r - showing the CTRA_TA windows");
	end;
end;

--[[
	============================================================
	CTRA_TA_Usage_Information

	Print a usage message to the chat log
]]
function CTRA_TA_Usage_Information()
    ctrata_chat("|cc0ffff00" .. CTRA_TA_VERSION_TEXT .. "|r Usage:");
    ctrata_chat("|cc0ffff00/ctrata show|r - show the CTRA_TA dialog");
    ctrata_chat("|cc0ffff00/ctrata hide|r - hide the CTRA_TA dialog");
    ctrata_chat("|cc0ffff00/ctrata toggle|r - toggle the CTRA_TA dialog");
    ctrata_chat("|cc0ffff00/ctrata position left|r - display the CTRA_TA dialog to the left of the MT windows");
    ctrata_chat("|cc0ffff00/ctrata position right|r - display the CTRA_TA dialog to the right of the MT windows");
    ctrata_chat("|cc0ffff00/ctrata position auto|r - display the CTRA_TA dialog to the side of the MT windows with more space");
    ctrata_chat("|cc0ffff00/ctrata version|r - display the version number");
end;


--[[
	============================================================
 	CTRA_TA_OnUpdate

	Update function!  Called on every frame :/

	We try to limit updates to every 0.25s (4 updates a second).

	Every time we update, we get the list of CTRA tanks, convert
	them to a list of unitids, and update our windows, including
	showing or hiding windows and repositioning.

]]
function CTRA_TA_OnUpdate(elapsed)
	if (this.durationTick == nil) then
		this.durationTick = 0;
	end;

	this.durationTick = this.durationTick + elapsed;

	-- Every 250ms, we update the frames
	if (this.durationTick > 0.25) then
		-- CT_RA_MainTanks is a list of player names... we want ids.
		if CT_RA_MainTanks then
			local tanklist = CTRA_TA_GetUnitList(CT_RA_MainTanks);
	
			ctrata_debug(3, "We have "..table.getn(tanklist or {}).." tanks.");
			CTRA_TA_SetWindows(tanklist);
	
			-- get rid of any excess ticks
			while (this.durationTick > 0.25) do
				this.durationTick = this.durationTick - 0.25;
			end;
		end;
	end;
end;


--[[
	============================================================
	CTRA_TA_GetUnitList

	Given a list of names (e.g. {[1]="Harna", [3]="Fred", [4]="Joe"}), return
	a list of corresponding unitids for the names (e.g. {[1]="raid13",
	[3]="raid23", [4]="raid39"}

]]
function CTRA_TA_GetUnitList(playernames)
	local unitlist = {};
	local added = 0;
	for k, v in pairs(playernames) do
		ctrata_debug(4, "getting unitid for "..v);
		for i = 1, GetNumRaidMembers() do
			if (UnitName("raid" .. i) == v ) then
				ctrata_debug(4, "unit id for "..v.." is raid"..i);
				unitlist[k] = "raid"..i;
				added = added + 1;
				break;
			end
		end
	end;

--	table.setn(unitlist, added);


	return unitlist;
end;


--[[
	============================================================
	CTRA_TA_SetWindows

	Given a list of MTs, show or hide each window based on
	whether there are any duplicate targets found.  The warning
	will be displayed next to the tank with aggro, or to the
	lowest numbered tank (MT1 in preference to MT2 etc) if none
	of the tanks has agro.

	The parameter is a list of unitids (raid13, raid4 etc)
	corresponding to the CTRA MT tank list.  There may be gaps in
	this list (e.g. {[3] = "raid13", [5] = "raid19"}, but it is
	expected that the list will be iterated in numerical order.

	Examples:

	 - 	If we have two tanks (A and B), and they are both on target T,
		then if A has aggro (Ttarget = A) we put "A, B" next to A,
		or "B, A" next to B if B has aggro, or "A, B" next to A if
		C (a non tank) has aggro (A is considered higher priority)

	 - 	If we have two tanks (A and B), and two targets TA and TB,
		we don't display anything if A is on TA and B is on TB, since
		CTRA will have changed the tank frame red anyway if TA and TB
		have agro on other people.

	 - 	If we have three tanks (A,B,C), and two targets TA and TB,
		and if A has target TA, B and C target TB, what we display
		depends on the agro of the targets.  A will have nothing, and
		B will have "B,C" if A or B has agro, and C will have "C, B" if
		C has agro.

	We format the outputs as follows:
	 - 	first tank is in green (it's their target)
	 -  subsequent tanks are yellow, unless that tank is the
	 -  current player, in which case we color it red.  We also color
		the background red.

]]
function CTRA_TA_SetWindows(tanklist)

	local targetlists = {}; -- list of lists of tanks on the same target
	local warnlist = {};	-- list of warnings to display
	local reverselist = {};	-- map from tank unitid to tank index

	-- Make a reverse lookup
	for k,v in pairs(tanklist) do
		reverselist[v] = k;
	end;

	-- Partition the target lists
	targetlists = CTRA_TA_CheckTargets(tanklist);
	
	--[[
		Now targetslist will be a list of lists of tanks with the same
		target.  For each list, we figure out which tank to assign
		the message to.  No tank should appear in two lists.

		We basically get up to n lists of up to n tanks on the same target,
		with n total list elements. Then, we decide which tank in each list
		gets the warning label.
	]]

	for k,targetgroup in pairs(targetlists) do
	
		ctrata_debug(4, "Target group "..k.." had "..table.getn(targetgroup).." members");

		if (table.getn(targetgroup) > 1) then

			-- The default warning is to appear next the first tank in the list
			local warningtarget = targetgroup[1];
			local msg = "";
			local bgc = 0;	-- 0 for yellow, 1 for red
	
			-- targetgroup should be a list of tanks on the same target
			for k2, tank in pairs(targetgroup) do    
				-- tank will be a particular unitid (e.g. raid12)
				-- See if this unit has aggro
				if (UnitIsUnit(tank.."targettarget", tank)) then
					-- Yep, so we put the warning next to this unit
					ctrata_debug(4, "Tank "..tank.." had aggro");
					warningtarget = tank;
					break;
				end;
			end;
	
			-- Now construct the warning string
			msg = "|cff80ff80" .. UnitName(warningtarget) .. "|r";
	
			for k2, tank in pairs(targetgroup) do
				if (tank ~= warningtarget) then
					if (UnitIsUnit(tank, "player")) then
						msg = msg .. ", |cffff8080" .. UnitName("player") .. "|r";
						bgc = 1; -- Use a red bg
					else
						msg = msg .. ", " .. UnitName(tank);
					end;
				end;
			end;
	
			-- And assign it!        
			warnlist[reverselist[warningtarget]] = {msg=msg, bgc=bgc};      
		end;
	end;

	-- Finally, show or hide the windows at the right spot
	local anchorsuffix = "LEFT";
	local relanchorsuffix = "RIGHT";
	local elem = getglobal("CT_RAMTGroup");
	local scale = elem:GetScale(); -- the CTRA scale setting changes this
	if (scale == 0) then
		scale = 0.001;
	end;

	-- Set our scale (each window in CTRA gets a scale too..)
	CTRA_TA_Window:SetScale(scale);

	-- Figure out the position option
	if(CTRA_TA_Options.position == "left") then
		anchorsuffix = "RIGHT";
		relanchorsuffix = "LEFT";
	elseif (CTRA_TA_Options.position == "auto") then
		-- Rough check to see if we should be on left or right
		local leftspace = elem:GetLeft();
		local rightspace = (UIParent:GetRight()/scale) - elem:GetRight();

		if (leftspace > rightspace) then
			anchorsuffix = "RIGHT";
			relanchorsuffix = "LEFT";
		end;
	end;
	
	-- Empty MT target are removed by CTRA, so make sure that we match
	-- the order of the MT targets
	local wl = {}
	local wlnr = 1;
	for i = 1, 10 do
		if (warnlist[i] ~= nil) then
			wl[wlnr]=warnlist[i];
		else
			if (CT_RA_MainTanks[i]~=nil) then
				wlnr = wlnr + 1;
			end;
		end;
	end;


	-- Now adjust each frame
	local framename;
	-- Are MT Target windows shown
	local mttShown = CT_RAMenu_Options["temp"]["ShowMTTT"];
	for i = 1, 10 do
		framename = "CTRA_TA_Tank"..i;
		local frame = getglobal(framename);
		

		if (wl[i] ~= nil) then
			ctrata_debug(3, "Showing warning '"..wl[i].msg.."' next to MT"..i);
			getglobal(framename .. "Text"):SetText(wl[i].msg);
	
			-- We align things like the CTRA tank frames
			frame:Hide();
			frame:ClearAllPoints();

			-- Set the scale here too (CTRA does, makes the borders fatter)
			frame:SetScale(scale);

			-- Make it the same height
			if (mttShown and anchorsuffix=="LEFT") then 
				ctrata_debug(3, "Aligning to MT Target window, anchorsuffix="..anchorsuffix);
				frame:SetPoint("TOP"..anchorsuffix, "CT_RAMTTGroupUnitButton"..i, "TOP"..relanchorsuffix);
				frame:SetPoint("BOTTOM"..anchorsuffix, "CT_RAMTTGroupUnitButton"..i, "BOTTOM"..relanchorsuffix); 
			else 
				ctrata_debug(3, "Aligning to MT window, anchorsuffix="..anchorsuffix);
				frame:SetPoint("TOP"..anchorsuffix, "CT_RAMTGroupUnitButton"..i, "TOP"..relanchorsuffix);
				frame:SetPoint("BOTTOM"..anchorsuffix, "CT_RAMTGroupUnitButton"..i, "BOTTOM"..relanchorsuffix); 
			end;

			-- Set the width correctly
			frame:SetWidth(150);

			-- Set the border based on CTRA
			if (CT_RAMenu_Options["temp"] and CT_RAMenu_Options["temp"]["HideBorder"]) then
				frame:SetBackdropBorderColor(1, 1, 1, 0);
			else
				frame:SetBackdropBorderColor(1, 1, 1, 1);
			end;

			-- Set the background color
			if (wl[i].bgc == 1) then
				frame:SetBackdropColor(1.0, 0.3, 0.3, 0.5);  -- red
			else
				frame:SetBackdropColor(1.0, 1.0, 0.3, 0.5);	 -- yellow
			end; 
			   
			frame:Show();             
		else                  
			frame:Hide();
		end;
	end;
end;


--[[
	============================================================
	CTRA_TA_CheckTargets

	Given a list of unitids, we return a list of groups of tanks on the
	same target.

	For example:

		Unitlist is {a, b, c, d, e}.
		Targets of these tanks are : {x, x, y, nil, z}

		We return:
		{
			{a, b}, -- all these units are on the same target (x)
			{c}, -- all these units (only one) are on the same target (y)
			{e}, -- same
		}

	If no unit in the list has a target, we just return {}
]]
function CTRA_TA_CheckTargets(unitlist)
	local samelist = {};
	local found = false;

	for k,v in pairs(unitlist) do
		ctrata_debug(3, "Checking targets for "..v);
		if (UnitExists(v.."target")) then
			-- See if this is the same as any of our existing target groups
			found = false;
			for k2,targetgroup in pairs(samelist) do
				if (UnitIsUnit(targetgroup[1].."target", v.."target")) then
					-- Same target as an existing targetgroup, add it to this group
					ctrata_debug(2, "Unit "..v.." has the same target as "..targetgroup[1]);
					table.insert(targetgroup, v);
					found = true;
				end;
			end;

			-- If we didn't find the current target, add a new target group with this tank
			if (found == false) then
				table.insert(samelist, {v});
			end;                  
		end;
	end;

	return samelist;
end;